#ifndef MIDI_BASE_H
#define MIDI_BASE_H
#include <cmath>
#include <iostream>
#include <map>
#include <algorithm>
#include <stack>
#include <cctype>
#include <allegro5\allegro5.h>
#include <allegro5/allegro_native_dialog.h>
#include "allegro5/allegro_audio.h"
#include "mfile/MidiFile.h"
#include "generation.h"

#define DEBUG

/* Base.h: Basic routines used by most parts of the program */

#define UNICHAR (char)-61

#define MILLISECONDS_TO_FRAMES(x) (x / 16.666)
#define FPS_TO_MS(x) (1.0 / x)

const double delay = BUFFERS * SAMPLES_PER_BUFFER / STREAM_FREQUENCY + 0.002;

extern std::string resourcePath;
extern int screenW, screenH;
extern int mouseX, mouseY;
extern bool buttonState;

inline bool linksklick() {
	return (buttonState ? !(buttonState = false) : false);
}

inline bool mouseover(int x1, int y1, int x2, int y2) {
	if(x1 > x2) return mouseover(x2, y1, x1, y2);
	if(y1 > y2) return mouseover(x1, y2, x2, y1);
	//std::cout << mouseX << "/" << mouseY << " " << x1 << "/" << y1 << " " << x2 << "/" << y2 << std::endl;
	return mouseX > x1 && mouseX < x2 && mouseY > y1 && mouseY < y2;
}

inline std::string getExtension(const std::string& filename) {
	auto it = filename.find_last_of('.');
	if(it == std::string::npos) return "";
	if(it+1 >= filename.length()) return "";
	
	std::string extension = filename.substr(it+1);
	
	for(int i = 0;i < extension.size();i++) {
		extension[i] = tolower(extension[i]);
	}
	
	return extension;
}

extern std::map<unsigned char, ALLEGRO_SAMPLE*> drumsamples;


class MidiEventListener {
	/* A Class that can listen to MIDI events. There's a list of all known listeners for the tracks
       and when they process a MIDI event, they call reactEvent on all registered listeners.
       
       When you want to listen to the MIDI Events you have to register your listener by calling
       
       Track::addListener(MidiEventListener*)
       (static function so no specific track object needed)
       
       If you call addListener on the same object more than one time, nothing will happen
    */
       
	public:
		virtual void reactEvent(int trackID, const MidiEvent& event) = 0;
};

inline float noteToFreq(unsigned char note) {
	/*
		Calculates the frequency of the MIDI note note (0-127)
		
		Based off http://www.sengpielaudio.com/calculator-notenames.htm
	*/
	return 440 * pow(2.0, (note-69.0)/12.0);
}



void abort_example(char const *format, ...);
void open_log(void);
void open_log_monospace(void);
void close_log(bool wait_for_user);
void log_printf(char const *format, ...);

inline const char *w2s(Waveform w) {
	/* Returns a name for the waveform w, used by the gui */
	switch(w) {
		case WAVEFORM_NONE:
			return "-";
			break;
		case WAVEFORM_SINE:
			return "~ Sine";
			break;
		case WAVEFORM_SQUARE:
			return "[] Square";
			break;
		case WAVEFORM_TRIANGLE:
			return "^^ Triangle";
			break;
		case WAVEFORM_SAWTOOTH:
			return "|\\|\\ Sawtooth";
			break;
	}
}

typedef unsigned int StringID;

class StringManager {
	/*
		Used for the communication between the main thread and the thread of the MIDI interpreter.
		
		Strings (for example filenames) get pushed here and the returned ID can be transmitted as a value
		for the Allegro User Event that the Interpreter recieves, who then pops the String.
		
		Stores the String in vec and the currently unused elements of vec in available. If no element is available,
		a new element is added to the vec and it's ID is added to "used".
		When a string gets popped, the corresponding ID is removed from used and added to available.
	*/
	private:
		std::stack<StringID> available;
		std::vector<StringID> used;
		
		std::vector<std::string> vec;
	public:
		StringID push(const std::string& str) {
			/*	Versucht ein bereits vorhandenes Element zu nehmen, wenn keine da sind, tut es das in den
				Vektor und markiert die Stelle als belegt
			*/
				
			if(available.empty()) {
				vec.push_back(str);
				used.push_back(vec.size()-1);
				return vec.size()-1;
			}
			StringID id = available.top();
			available.pop();
			used.push_back(id);
			vec[id] = str;
			return id;
		}
		
		const std::string& pop(StringID id) {
			auto it = std::find(used.begin(), used.end(), id);
			used.erase(it);
			
			available.push(id);
			
			return vec[id];
		}
};

extern StringManager *stringManager;

/* The Events that the Program uses to communicate with the Interpreter Thread */
#define EVENT_FILE_ADD   ALLEGRO_GET_EVENT_TYPE('B','E','T','T')
#define EVENT_NEXT  (EVENT_FILE_ADD + 1)
#define EVENT_PLAY      (EVENT_FILE_ADD + 2)
#define EVENT_STOP      (EVENT_FILE_ADD + 3)
#define EVENT_LAST		(EVENT_FILE_ADD + 4)
#define EVENT_PREV		(EVENT_FILE_ADD + 5)

/* The Different Meta Events that MIDI has & get to the registered MidiEventListeners when a Track 
  gets to them. 
  
  "Meta events are used for special non-MIDI events (...). They have the general form : 0xFF type length data
  
  	type specifies the type of Meta event (0 - 127).
	length is a variable length quantity (as used to represent delta times)
			specifying the number of bytes that make up the following data.
			Some Meta events do not have a data field, whereupon length is 0. "
			
  For a explanation of the different events look at:
  
  http://www.somascape.org/midi/tech/mfile.html#meta
  
  or
  
  http://www.music.mcgill.ca/~ich/classes/mumt306/StandardMIDIfileformat.html#BM3_1
  
*/
#define EV_SEQ_ID 0x00
#define EV_TEXT 0x01
#define EV_COPYRIGHT 0x02
#define EV_NAME 0x03
#define EV_INST_NAME 0x04
#define EV_LYRIC 0x05
#define EV_MARKER 0x06
#define EV_CUE_POINT 0x07
#define EV_CHANNEL_PREFIX 0x20
#define EV_TRACK_END 0x2F
#define EV_SET_TEMPO 0x51
#define EV_SMPTE_OFFS 0x54
#define EV_TIME_SIGNATURE 0x58
#define EV_KEY_SIGNATURE 0x59
#define EV_SEQUENCER_SPEC 0x7F


inline void makeAllegroReady(std::string& str) {
	/* The Allegro Text functions fail when a string is not in UTF-8 and contains
		chars >127 (Umlauts and other special non-english characters for example)
		The solution here is to just replace them with '?')
	*/
	for(int i = 0;i < str.size();i++) {
		unsigned char c = str[i];
		
		if(c > 127) {
			str[i] = '?';
		}
	}
}

inline std::vector<std::string> readPlaylist(const std::string& filename) {
	/* Read in a m3u playlist and return the files as a vector */
	std::fstream file(filename, std::ios::in);
	std::vector<std::string> playlist;
	
	if(!file) {
		return playlist;
	}
	
	std::string line;
	while(std::getline(file, line)) {
		if(getExtension(line) == "mid" || getExtension(line) == "midi") {
			if(line.length() > 0 && line.find_first_not_of(' ') != '#') {
				playlist.push_back(line);
			}
		}
	}
	
	return playlist;
}

#endif
