#ifndef GENERATION_H
#define GENERATION_H
#include <allegro5\allegro5.h>
#include "allegro5/allegro_audio.h"

#define PI                    (ALLEGRO_PI)
#define TWOPI                 (2.0 * PI)
#define SAMPLES_PER_BUFFER    (1024)
#define STREAM_FREQUENCY      (44100)
#define BUFFERS               (2)

const double dt = 1.0 / STREAM_FREQUENCY;

enum Waveform {
    WAVEFORM_NONE,
    WAVEFORM_SINE,
    WAVEFORM_SQUARE,
    WAVEFORM_TRIANGLE,
    WAVEFORM_SAWTOOTH
};

static void sine(float *buf, size_t samples, double t,
                 float frequency, float phase) {
	const double w = TWOPI * frequency;
	unsigned i;

	for (i = 0; i < samples; i++) {
		double ti = t + i * dt;
		buf[i] = sin(w * ti + phase);
	}
}

static void square(float *buf, size_t samples, double t,
                   float frequency, float phase) {
	const double w = TWOPI * frequency;
	unsigned i;

	for (i = 0; i < samples; i++) {
		double ti = t + i * dt;
		double x = sin(w * ti + phase);

		buf[i] = (x >= 0.0) ? 1.0 : -1.0;
	}
}

static void triangle(float *buf, size_t samples, double t, float frequency, float phase) {
	const double w = TWOPI * frequency;
	unsigned i;

	for (i = 0; i < samples; i++) {
		double tx = w * (t + i * dt) + PI/2.0 + phase;
		double tu = fmod(tx/PI, 2.0);

		if (tu <= 1.0)
			buf[i] = (1.0 - 2.0 * tu);
		else
			buf[i] = (-1.0 + 2.0 * (tu - 1.0));
	}
}

static void sawtooth(float *buf, size_t samples, double t, float frequency, float phase) {
	const double w = TWOPI * frequency;
	unsigned i;

	for (i = 0; i < samples; i++) {
		double tx = w * (t + i * dt) + PI + phase;
		double tu = fmod(tx/PI, 2.0);

		buf[i] = (-1.0 + tu);
	}
}

static void generate_wave(Waveform type, float *buf, size_t samples, double t, float frequency, float phase) {
	if(frequency == 0.0 || type == WAVEFORM_NONE) {
		al_fill_silence(buf, samples, ALLEGRO_AUDIO_DEPTH_FLOAT32, ALLEGRO_CHANNEL_CONF_1);
		return;
	}
	
	switch (type) {
		case WAVEFORM_SINE:
			sine(buf, samples, t, frequency, phase);
			break;
		case WAVEFORM_SQUARE:
			square(buf, samples, t, frequency, phase);
			break;
		case WAVEFORM_TRIANGLE:
			triangle(buf, samples, t, frequency, phase);
			break;
		case WAVEFORM_SAWTOOTH:
			sawtooth(buf, samples, t, frequency, phase);
			break;
	}
}


#endif
