#include "interpreter.h"
#include "synth.h"

Interpreter::Interpreter() : timer(mtempoToSpeed(500000, 60)) {
	queue = al_create_event_queue();
	al_register_event_source(queue, timer.getEventSource());
}

Interpreter::~Interpreter() {
	tracks.clear();
	al_destroy_event_queue(queue);
}

std::string extractText(const MidiEvent& event) {
	unsigned char len = event[2];
			
	std::string str;
	for(int i = 0;i < len;i++) str += event[3+i];
	return str;
}

void Interpreter::reactEvent(int trackID, const MidiEvent& event) {
	if(event.isMeta()) {
		if(event[1] == EV_TEXT) {
			std::string text = extractText(event);
			std::cout << "[INFO] Text: \"" << text << "\"" << std::endl;
		} else if(event[1] == EV_COPYRIGHT) {
			std::string text = extractText(event);
			std::cout << "[INFO] (C) \"" << text << "\"" << std::endl;
			if(copyright != "") copyright += " ";
			copyright += text;
		} else if(event[1] == EV_NAME) {
			std::string name = extractText(event);
			
			if(channelPrefix != 128) trackID = channelPrefix;
			
			if(trackID == 0 && channelPrefix == 128) {
				if(fileSongName) {
					songName = "";
					fileSongName = false;
				}
				std::cout << "Song Name: \"" << name << "\"" << std::endl;
				if(songName != "") songName += " ";
				songName += name;
				makeAllegroReady(songName);
			} else {
				trackID -= 1;
				std::cout << "Channel " << trackID << " Name: \"" << name << "\"" << std::endl;
				if(trackID < 16 && trackID >= 0)synth->getVoice(trackID)->setName(name);
			}
		} else if(event[1] == EV_INST_NAME) {
			std::string name = extractText(event);
			
			if(channelPrefix < 128) trackID = channelPrefix;
			std::cout << "Instrument " << trackID << " Name: \"" << name << "\"" << std::endl;
		} else if(event[1] == EV_LYRIC) {
			std::string text = extractText(event);
			std::cout << "[INFO] '" << text << "'" << std::endl;
			std::cout << trackID << std::endl;
			if(lyrics.length() > 50 && (lyrics[lyrics.length()-1] == ' ' || lyrics[lyrics.length()-1] == '.' || lyrics[lyrics.length()-1] == ',')) lyrics = "";
			lyrics += text;
		} else if(event[1] == EV_MARKER) {
			std::string text = extractText(event);
			std::cout << "[INFO] Marker '" << text << "'" << std::endl;
		} else if(event[1] == EV_CUE_POINT) {
			std::string text = extractText(event);
			std::cout << "[INFO] Cue '" << text << "'" << std::endl;
		} else if(event[1] == EV_SET_TEMPO) {
			timer.setSpeed(mtempoToSpeed(decodeTempo(event), file.getTicksPerQuarterNote()));
		} else if(event[1] == EV_CHANNEL_PREFIX){
			channelPrefix  = event[3];
			std:cout << "[INFO] Channel Prefix: " << std::dec << (int) channelPrefix << std::endl;
		} else if(event[1] == EV_TRACK_END) {
			tracks[trackID].end();
			activeTracks--;
			if(activeTracks <= 0) this->next();
		} else {
			std::cout << "[INFO] Unhandled Event 0x" << std::hex << (int) event[1] << std::dec << std::endl;
		}
	} else {
		channelPrefix = 128;
	}
}

void Interpreter::reset() {
	for(int i = 0;i < tracks.size();i++) {
		tracks[i].reset();
	}
	
	fileSongName = true;
	tickPos = 0;
	channelPrefix = 128;
	copyright = "";
	lyrics = "";
	
	for(int i = 0;i < 16;i++) {
		synth->getVoice(i)->setName("");
	}
}

void Interpreter::openFile(const std::string& filename) {
	std::fstream test(filename, std::ios::in);
	if(!test) throw(std::string("Can't open file ") + filename);
	test.close();
	
	if(getExtension(filename) != "mid" && getExtension(filename) != "midi") throw(std::string("Unrecognized File Extension: ") + filename);
	
	gotSong = true;
	file.read(filename);
	file.absoluteTicks();
	
	tracks.clear();
	for(size_t i = 0;i < file.getTrackCount();i++) {
		tracks.push_back(Track(&file, i));
	}
	activeTracks = tracks.size();
	
	Track::addListener(this);
	
	#ifdef DEBUG
	std::cout << file.getTrackCount() << " Tracks" << std::endl;
	#endif
	
	auto pos = filename.find_last_of('\\');
	if(pos == std::string::npos) pos= filename.find_last_of('/');
	if(pos == std::string::npos) {
		songName = filename;
	} else {
		songName = filename.substr(pos+1);
	}
	
	if(1 < tracks.size()) {
		for(int i = 0;i < file[1].getSize();i++) {
			if(file[1][i].isMeta() && file[1][i][1] == EV_LYRIC) {
				std::string text = extractText(file[1][i]);
				completeLyrics += text;
			}
		}
	}


	for(int i = 1;i < tracks.size();i++) {
		if(tracks[i].getLength() > tracks[longestTrack].getLength()) longestTrack = i;
	}

	this->reset();
}

void Interpreter::check() {
	std::string filename;	
	
	ALLEGRO_EVENT ev;
	while(al_get_next_event(queue, &ev)) {
		switch(ev.type) {
			case EVENT_PLAY:
				if(!this->hasSong()) {
		 			al_show_native_message_box(nullptr, "Error", "No Song Loaded!", "There was no MIDI-file opened before, so nothing could be played!", "I Guess", ALLEGRO_MESSAGEBOX_ERROR);
				} else {
					this->reset();
		 			this->play();
				}
				break;
			case EVENT_STOP:
				this->stop();
				synth->panic();
				break;
			case EVENT_NEXT:
				this->next();
				if(!this->isFinished()) {
					this->stop();
					synth->panic();
				}
				break;
			case EVENT_PREV:
				this->prev();
				if(!this->isFinished()) {
					this->stop();
					synth->panic();
				}
				break;
			case EVENT_LAST:
				std::cout << "[INFO] Skipping to last!" << std::endl;
				playlist.pos = playlist.data.size()-1;
				if(!this->isFinished()) {
					this->stop();
					synth->panic();
				}
				break;
			case EVENT_FILE_ADD:
				if(this->isFinished()) {
					if(!this->isFinished()) {
						this->stop();
						synth->panic();
					}
				}
				
				std::string filename = stringManager->pop(ev.user.data1);
				
				playlist.data.push_back(filename);
				std::cout << "[INFO] Adding " << filename << " to Playlist" << std::endl;
				break;
		}
		
	}
	
	if(this->isFinished()) {
		if(playlist.pos >= playlist.data.size()) return;
		
		this->openFile(playlist.data[playlist.pos]);
		synth->panic();
		this->play();
	}
	
	while(timer.getCount() > 0) {
		timer.addCount(-1);
		tickPos++;
		
		for(size_t i = 0;i < tracks.size();i++) {
			tracks[i].update(tickPos);
		}
	}
}
